#!/usr/bin/env python

# Copyright (C) 2012 Robin Trew

# Permission is hereby granted, free of charge, 
# to any person obtaining a copy of this software 
# and associated documentation files (the "Software"), 
# to deal in the Software without restriction, 
# including without limitation the rights to use, copy, 
# modify, merge, publish, distribute, sublicense, 
# and/or sell copies of the Software, and to permit persons 
# to whom the Software is furnished to do so, 
# subject to the following conditions:

# The above copyright notice and this permission notice 
# shall be included in ALL copies 
# or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, 
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
# OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""Read OmniFocus Sqlite to indented text outline"""
# ver .007

import os, sys, sqlite3, re, codecs
from xml.sax.saxutils import quoteattr

STR_AUTHOR = "Robin Trew 2012"
STR_PUB = "Orginally published on http://forums.omnigroup.com"

# position of fields in the nodes
FLD_PARENT_ID = 0
FLD_ID = 1
FLD_NAME = 2
FLD_NOTE = 3
FLD_CHILN = 4
INT_ALL_FIELDS = 5

# types of OmniFocus object
OF_INBOX = 0
OF_FOLDER = 1
OF_PROJECT = 2
OF_TASK = 3

# opml fragments
OPML_NODE_START = '<outline '
OPML_LEAF_CLOSE = '/>'
OPML_PARENT_CLOSE = '</outline>'
OPML_HEAD_TO_EXPAND = '<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n\
<opml version=\"1.0\">\n\
<head>\n    <title>OmniFocus Active Tasks</title>\n\
    <expansionState>'
OPML_HEAD_FROM_EXPAND = '</expansionState>\n\
</head>\n\
<body>\n'
OPML_TAIL = ' \
    </body></opml>'

def main():
    """Read from OmniFocus cache then write utf8-encoded opml file"""
    flat_of = get_flat_omnifocus()
    lst_clean = clean_and_escape(list(flat_of))
    dict_parents = get_parent_posns(lst_clean)
    flat_to_nested(lst_clean, dict_parents)
    lst_clean = [[None, None, '\"Focus\"', None, lst_clean]]
    
    # Optionally include links of the form omnifocus:///task/o49Lh0LWVjA
    # (look for a 0 or 1 as the second argument to the call)
    try:
        bln_add_links = int(sys.argv[2])
    except IndexError:
        bln_add_links = 0
        
    str_opml = opml_doc(lst_clean, bln_add_links)
    
    # Write out the specified file (or print to screen for debugging)
    try:
        str_out_path = sys.argv[1]
        tmp_file = codecs.open(str_out_path, "w", "utf-8")
        tmp_file.write(str_opml)
        tmp_file.close()
    except IndexError:
        print str_opml # for debugging
     
def opml_doc(lst, bln_links):
    """Assemble an OPML document"""
    return OPML_HEAD_TO_EXPAND + '0' + OPML_HEAD_FROM_EXPAND + \
    tree_to_opml(lst, '\t', bln_links) + OPML_TAIL
        
def tree_to_opml(lst, str_indent, bln_links):
    """Traverse tree to build nested opml outline"""
    str_out = ""
    str_deeper = str_indent + "\t"
    for peer in lst:
        [str_id, str_name, str_note] = peer[FLD_ID:FLD_CHILN]
        str_out = str_out + OPML_NODE_START + attr('text', str_name)
        if (bln_links != 0) and (str_id is not None):
            str_out = str_out + of_link(str_id)
        if (str_note is not None):
            str_out = str_out + attr('_note', peer[FLD_NOTE])
        
        if (len(peer) >= INT_ALL_FIELDS):
            str_out = str_out + '>\n' + str_indent + \
                tree_to_opml(peer[FLD_CHILN], str_deeper, bln_links) + \
                '\n' + str_indent + OPML_PARENT_CLOSE
        else:
            str_out = '%s%s\n' % (str_out, OPML_LEAF_CLOSE)
    return str_out

def of_link(str_id):
    """Build a LINK attribute from an OmniFocus id"""
    return '%s=\"%s%s\" ' % ('LINK', 'omnifocus:///task/', str_id)
    
def attr(str_name, str_value):
    """Build OPML attribute pairings"""
    return '%s=%s '  % (str_name, str_value)
    
def get_flat_omnifocus():
    """Read Omnifocus cache data into flat ordered list"""
    # Where is the ofoc cache ?
    ofoc = os.path.expanduser('~/Library/Caches/com.omnigroup.OmniFocus')
    if not os.path.exists(ofoc):
        ofoc = ofoc + '.MacAppStore'

    cache_db = ofoc +  '/OmniFocusDatabase2'

    # Prepare to query the ofoc cache
    con = sqlite3.connect(cache_db)
    csr = con.cursor()

    # Set up a temporary table reducing folders projects
    # and tasks to one record type (including parent)
    csr.execute('CREATE TEMP TABLE fpt_active (objtype integer, parent text, \
        persistentIdentifier text, name text, noteXMLData blob, rank integer)')
    
    # SPECIFY THE DATA
    # Gather active folders,
    tpl = (OF_FOLDER, 1)
    csr.execute('INSERT INTO fpt_active SELECT ?, \
        parent, persistentIdentifier, \
        name, noteXMLData, rank FROM Folder WHERE effectiveActive = ?', tpl)
    # active projects,
    tpl = (OF_PROJECT, 'active', 1, None)
    csr.execute('INSERT INTO fpt_active SELECT ?, p.folder, \
        tpl.persistentIdentifier, tpl.name, tpl.noteXMLData, rank \
        FROM (task tpl join projectinfo p on tpl.projectinfo = p.pk) WHERE p.status = ? \
         and p.folderEffectiveActive = ? and tpl.dateCompleted is ?', tpl)
    # and uncompleted tasks in those active projects
    tpl = (OF_TASK, 0, None, None)
    csr.execute('INSERT INTO fpt_active SELECT ?, parent, \
        persistentIdentifier, name, noteXMLData, rank \
    FROM task WHERE inInbox = ? and projectInfo is ? and effectiveContainingProjectInfoActive \
        and dateCompleted is ?', tpl)
    # An invented INBOX root node
    tpl = (OF_INBOX, None, 'Inbox', 'Inbox', None, None)
    csr.execute('INSERT INTO fpt_active SELECT ?, ?, ?, ?, ?, ?', tpl)
    # and all the inbox tasks
    tpl = (OF_TASK, 1, None, None)
    csr.execute('INSERT INTO fpt_active SELECT ?, \"Inbox\", \
        persistentIdentifier, name, noteXMLData, rank \
    FROM task WHERE inInbox = ? and projectInfo is ? and effectiveContainingProjectInfoActive \
        and dateCompleted is ?', tpl)
    
    # Set up recursive building of temporary table to get nested folder order
    csr.execute('CREATE TEMP TABLE Nested_Order (objtype integer, parent text, \
        persistentIdentifier text, name text, noteXMLData blob)')
    csr.execute('PRAGMA recursive_triggers = TRUE')
    csr.execute('CREATE TEMP TRIGGER order_Folders AFTER INSERT \
    ON Nested_Order \
    BEGIN \
        INSERT INTO Nested_Order SELECT objtype, parent, persistentIdentifier, \
            name, noteXMLData FROM fpt_active \
        WHERE fpt_active.parent = new.persistentIdentifier ORDER BY rank; \
    END')

    # Launch recursive process from top level objects
    # to build a temporary ordered table
    tpl = (None, )
    csr.execute('INSERT INTO Nested_Order SELECT objtype, parent, \
        persistentIdentifier, name, noteXMLData \
    FROM fpt_active WHERE parent is ? ORDER BY rank', tpl)
    # objtype!= 3 and
    # Harvest the recursively ordered temporary table

    csr.execute('SELECT parent, persistentIdentifier, name, \
        CAST(noteXMLData as text) FROM Nested_Order')
    flat_of = csr.fetchall()
    csr.execute('DROP TABLE fpt_active')
    csr.execute('DROP TABLE Nested_Order')
    return flat_of

def flat_to_nested(lst, dict_parents):
    """Bottom-up folding of ordered list into nested list"""

    for i in reversed(xrange(len(lst))):
        lst_node = lst[i]
        id_parent = lst_node[FLD_PARENT_ID]
        if (id_parent is not None):
            lst_parent = lst[dict_parents[id_parent]]
            # Move this node to the start of its parent's
            # children, creating a child list if needed
            if (len(lst_parent) < INT_ALL_FIELDS):
                lst_parent.append([lst_node])
            else:
                lst_parent[FLD_CHILN].insert(0, lst_node)
            lst.pop(i)

def get_parent_posns(lst):
    """Build a dictionary of the parent locations"""
    
    set_parent = set()
    dict_parent = {}
    # from bottom to top of the ordered node list
    for i in reversed(xrange(len(lst))):
        lst_node = lst[i]
        id_node = lst_node[FLD_ID]
        if (id_node in set_parent):
            dict_parent[id_node] = i 
        set_parent.add(lst_node[FLD_PARENT_ID])            
    return dict_parent

# Convert the captured flat_of from tuples to list format,
# convert the notes from XML to text,
# and escape both notes and text
def clean_and_escape(lst):
    """Clean notes and escape text"""
    # Match a link value pair
    xml_link = re.compile('<value\\ key=\\"link\\">[^a-z/]*([^<]*)</value>')
    # Match remaining value pairs
    xml_value = re.compile('<value.*?value>')
    # Match remaining tags
    xml_tag = re.compile('<.*?>')
    keep_links = xml_link.sub
    strip_values = xml_value.sub
    strip_tags = xml_tag.sub
    for i in range(len(lst)):
        lst_node = list(lst[i])
        lst_node[FLD_NAME] = quoteattr(lst_node[FLD_NAME])
        if lst_node[FLD_NOTE] is not None:
            lst_node[FLD_NOTE] = quoteattr(strip_tags('', strip_values('', \
              keep_links(' \\1 ', lst_node[FLD_NOTE]))))
        lst[i]=lst_node
    return lst

main()










